"""
Main web service CLI

"""
from click import command, option
from microcosm.loaders import load_from_dict
from microcosm.object_graph import ObjectGraph

from microcosm_sagemaker.app_hooks import create_serve_app
from microcosm_sagemaker.commands.options import input_artifact_option
from microcosm_sagemaker.profiling import enable_profiling


@command()
@option(
    "--host",
    default="127.0.0.1",
)
@option(
    "--port",
    type=int,
)
@option(
    "--debug/--no-debug",
    default=False,
)
@option(
    "--profile/--no-profile",
    default=False,
)
@input_artifact_option()
def main(host, port, debug, profile, input_artifact):
    graph = create_serve_app(
        debug=debug,
        extra_loader=load_from_dict(
            root_input_artifact_path=input_artifact.path,
        ),
    )

    if profile:
        enable_profiling(graph)

    run_serve(
        graph=graph,
        host=host,
        port=port,
    )


def run_serve(
    graph: ObjectGraph,
    host: str,
    port: int,
) -> None:
    graph.flask.run(
        host=host,
        port=port or graph.config.flask.port,
    )



from dataclasses import dataclass
from urllib.parse import urlparse

from boto3 import client
from botocore.exceptions import ClientError


@dataclass
class S3Object:
    bucket: str
    key: str

    @property
    def path(self) -> str:
        return f"s3://{self.bucket}/{self.key}"

    @property
    def exists(self) -> bool:
        try:
            s3 = client("s3")
            s3.head_object(Bucket=self.bucket, Key=self.key)
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return False
            else:
                raise
        else:
            return True

    @classmethod
    def from_url(cls, url):
        parsed = urlparse(url)
        return cls(bucket=parsed.netloc, key=parsed.path.strip("/"))



import os

from microcosm.api import binding, defaults
from microcosm_logging.decorators import logger

from microcosm_sagemaker.decorators import metrics_observer


try:
    import wandb
    from wandb.apis import CommError
except ImportError:
    pass


@metrics_observer()
@logger
@binding("weights_and_biases")
@defaults(
    enable=None,
)
class WeightsAndBiases:
    def __init__(self, graph):
        self.graph = graph
        self.enable = (
            # If it is not explicitly enabled or disabled, enable if not testing.
            graph.config.weights_and_biases.enable or not graph.metadata.testing
        )
        self.project_name = graph.metadata.name.replace("_", "-")
        self.bundle_and_dependencies_config_extractor = self.graph.bundle_and_dependencies_config_extractor
        self.active_bundle = getattr(graph, graph.config.active_bundle)
        self.run_path = getattr(graph.config.wandb, "run_path", None)
        self.wandb_run = None

    def init(self):
        # Sometimes, the entity in the wandb run path appears as None.
        # To make sure this does not happen, we explicityly set the wandb entity.
        # https://docs.wandb.com/library/environment-variables
        os.environ["WANDB_ENTITY"] = "globality"

        # Only initialize wandb if it is enabled
        if not self.enable:
            return

        # Only initialize/load wandb if it is not already created/loaded
        if self.wandb_run is not None:
            return

        # Pushing into an existing wandb experiment
        if self.run_path:

            try:
                self.wandb_run = wandb.Api().run(path=self.run_path)
                self.logger.info(f"The existing Weights & Biases run was loaded: {self.run_path}")

            except CommError:
                # If the previous run cannot be found:
                #   - Start a new run instead
                #   - Add a note to the new run that this run is related to a previous run
                self.logger.warning(f"Could not find run: {self.run_path}")
                self.wandb_run = self._create_new_wandb_run(
                    notes=f"Related to previous run: {self.run_path}",
                )

        # Creating a new wandb experiment
        else:
            self.wandb_run = self._create_new_wandb_run()

    def _create_new_wandb_run(self, notes=None):
        wandb_run = wandb.init(
            project=self.project_name,
            config=self.bundle_and_dependencies_config_extractor(self.active_bundle),
            notes=notes,
        )

        # Injecting the wandb run path into the config
        self.graph.config.wandb.run_path = wandb_run.path
        # Setting the run_path for the current instance.
        # This prevents a new wandb instantiation if the `init` method is called again.
        self.run_path = wandb_run.path

        # Adding the link to the Weights & Biases run to the landing page
        landing_convention_links = self.graph.config.landing_convention.get("links", {})
        landing_convention_links.update({"Weights & Biases": wandb_run.get_url()})
        self.graph.config.landing_convention.update({"links": landing_convention_links})

        self.logger.info(f"A new `weights & biases` run was created: {wandb_run.path}")

        return wandb_run

    def log_timeseries(self, **kwargs):
        step = kwargs.pop("step", None)
        self.wandb_run.log(kwargs, step=step)
        return None

    def log_static(self, **kwargs):
        self.wandb_run.summary.update(kwargs)
        return None



from io import BytesIO
from pathlib import Path
from typing import (
    Any,
    List,
    Mapping,
    Optional,
    Union,
)

import numpy as np
from hamcrest.core.base_matcher import BaseMatcher
from hamcrest.core.string_description import StringDescription
from sklearn.externals import joblib

from microcosm_sagemaker.testing.bytes_extractor import ExtractorMatcherPair


class IsObjectEqual(BaseMatcher):
    """
    Matcher to compare the object A with another object B.
    Sometimes there is some randomness within these generated files, so the user can optionally
    provide a `ignore_properties` array which will ignore mismatches of those given object
    properties.  For instance, if we have:
    MyObject(foo=bar, extras=Test)
    MyObject(foo=bar, extras=Test2)
    If ignore_properties is not specified, these will be unequal.  If ignore_properties is
    set to ["extras"], they will be equal.

    """
    def __init__(self, obj: Any, ignore_properties: List = []) -> None:
        self.ignore_properties = ignore_properties
        self.compare_object = obj

    def _matches(self, item: Any) -> bool:
        compare_a = self._get_dict_structure(self.compare_object)
        compare_b = self._get_dict_structure(item)

        total_attributes = (
            attr
            for attr in set(compare_a.keys()) | set(compare_b.keys())
            if attr not in self.ignore_properties
        )

        for attr in total_attributes:

            if attr not in compare_a or attr not in compare_b:
                return False

            if type(compare_a[attr]) != type(compare_b[attr]):  # noqa: 721
                return False

            if isinstance(compare_a[attr], np.ndarray):
                if (compare_a[attr] != compare_b[attr]).any():
                    return False
            elif compare_a[attr] != compare_b[attr]:
                return False

        return True

    def describe_to(self, description: StringDescription) -> None:
        description.append_text(str(self.compare_object))

    def _get_dict_structure(self, obj: Any) -> Mapping[str, Any]:
        if isinstance(obj, dict):
            return obj
        # the matcher expects a dictionary in the _matches method
        # -> we convert the set into a dictionary with a dummy key
        elif isinstance(obj, set):
            return dict(dummy_attrb=obj)
        return vars(obj)


def pickle_extractor(binary: bytes) -> Any:
    return joblib.load(BytesIO(binary))


def matches_object(
    obj: Union[object, dict],
    ignore_properties: Optional[List[str]] = None,
) -> IsObjectEqual:
    if ignore_properties is None:
        ignore_properties = []
    return IsObjectEqual(obj, ignore_properties)


def matches_with_object(obj: Any) -> IsObjectEqual:
    return matches_object(obj)


def _get_dir_pickles(dir: Path, pickle_suffixes: List[str]) -> List[Path]:
    return [
        subpath.relative_to(dir)
        for subpath in dir.glob("**/*")
        if subpath.suffix in pickle_suffixes
    ]


def _create_matchers(dir_pickles: List[Path]) -> Mapping[Path, ExtractorMatcherPair]:
    return {
        path: ExtractorMatcherPair(
            pickle_extractor,
            matches_with_object,
        )
        for path in dir_pickles
    }


def create_pickle_matchers_for_dir(dir, pickle_suffixes=[".pk", ".pickle", ".pkl"]):
    dir_pickles = _get_dir_pickles(dir, pickle_suffixes)
    matchers = _create_matchers(dir_pickles)
    return matchers



"""
SageMaker global constants.

"""
from pathlib import Path


APP_HOOKS_GROUP = "microcosm_sagemaker.app_hooks"

EVALUATE_APP_HOOK = "evaluate"
SERVE_APP_HOOK = "serve"
TRAIN_APP_HOOK = "train"

SAGEMAKER_PREFIX = Path("/opt/ml/")
ARTIFACT_CONFIGURATION_PATH = "configuration.json"


class SagemakerPath:
    INPUT_DATA = SAGEMAKER_PREFIX / "input/data"
    MODEL = SAGEMAKER_PREFIX / "model"
    OUTPUT = SAGEMAKER_PREFIX / "output"
    HYPERPARAMETERS = SAGEMAKER_PREFIX / "input/config/hyperparameters.json"


# NOTE: We can only use up to 10 dimensions:
# https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudwatch.html
CLOUDWATCH_MAX_DIMENSIONS = 10



from dataclasses import dataclass, field

from hamcrest import assert_that
from hamcrest.core.base_matcher import BaseMatcher

from microcosm_sagemaker.bundle import Bundle


@dataclass
class BundlePredictionCheck:
    return_value_matcher: BaseMatcher
    args: list = field(default_factory=list)
    kwargs: dict = field(default_factory=dict)

    def check_bundle(self, bundle: Bundle) -> None:
        assert_that(
            bundle.predict(*self.args, **self.kwargs),
            self.return_value_matcher,
        )



from pathlib import Path

from microcosm_sagemaker.commands import evaluate
from microcosm_sagemaker.testing.cli_test_case import CliTestCase


class EvaluateCliTestCase(CliTestCase):
    """
    Helper base class for writing tests of the evaluate cli.

    """
    input_data_path: Path
    input_artifact_path: Path

    def test_evaluate(self) -> None:
        """
        Runs the `evaluate` command on the given `input_data_path` and
        `input_artifact_path`.
        """
        self.run_and_check(
            command_name="evaluate",
            command=evaluate.main,
            args=[
                "--input-data",
                str(self.input_data_path),
                "--input-artifact",
                str(self.input_artifact_path),
            ],
        )



from microcosm.config.model import Configuration
from microcosm.loaders import expand_config
from microcosm.metadata import Metadata

from microcosm_sagemaker.constants import SagemakerPath


def load_default_microcosm_runserver_config(metadata: Metadata) -> Configuration:
    """
    Construct runserver default configuration.

    """

    config = Configuration(
        root_input_artifact_path=SagemakerPath.MODEL,
    )

    return config


def split_args_into_key_values(args):
    """
    Receives a list of string command line arguments.
    For the ones with the `--KEY=VALUE` format (which is the format used by wandb to inject hyperparameters),
    creates a {KEY: VALUE} dictionary and returns.

    """
    args = [
        arg.strip()[2:]
        for arg in args
        if arg.strip().startswith("--") and "=" in arg
    ]

    return {
        arg.split("=")[0]: arg.split("=")[1]
        for arg in args
    }


def value_func_with_quote_handling(value):
    """
    To pass values that include spaces through command line arguments, they must be quoted. For instance:
    --parameter='some value with space' or
    --parameter="some value with space".
    In this value parser, we remove the quotes to receive a clean string.

    """
    if value[0] == value[-1] and value[0] in ["'", '"']:
        return value[1:-1]
    return value


def load_config_from_command_line_arguments(command_line_config):
    """
    Load configuration from command line arguments.

    """
    config = expand_config(
        split_args_into_key_values(command_line_config),
        value_func=value_func_with_quote_handling
    )

    return Configuration(config)



from click import Path, option

from microcosm_sagemaker.artifact import RootInputArtifact, RootOutputArtifact
from microcosm_sagemaker.constants import SagemakerPath
from microcosm_sagemaker.input_data import InputData


def input_data_option():
    return option(
        "--input-data",
        type=Path(
            resolve_path=True,
            file_okay=False,
            exists=True,
        ),
        callback=_make_click_callback(InputData),
        default=SagemakerPath.INPUT_DATA,
        help="Path of the folder that houses the train/test datasets",
    )


def input_artifact_option():
    return option(
        "--input-artifact",
        type=Path(
            resolve_path=True,
            file_okay=False,
            exists=True,
        ),
        callback=_make_click_callback(RootInputArtifact),
        default=SagemakerPath.MODEL,
        help="Path from which to load artifact",
    )


def output_artifact_option():
    return option(
        "--output-artifact",
        type=Path(
            resolve_path=True,
            file_okay=False,
            writable=True,
        ),
        callback=_make_click_callback(RootOutputArtifact),
        default=SagemakerPath.MODEL,
        help="Path for outputting trained artifact",
    )


def _make_click_callback(function):
    """
    Given a `function`, returns a callback function that can be used for a
    click option's `callback=` to apply `function` to the value before passing
    the argument to the command

    """
    def callback(ctx, param, value):
        return function(value)

    return callback



from abc import ABC
from pathlib import Path

from microcosm.loaders import load_each, load_from_dict

from microcosm_sagemaker.app_hooks import create_evaluate_app
from microcosm_sagemaker.bundle import Bundle
from microcosm_sagemaker.evaluation import Evaluation
from microcosm_sagemaker.input_data import InputData


class EvaluationTestCase(ABC):
    # These should be defined in actual test case derived class
    evaluation_name: str
    root_input_artifact_path: Path
    input_data_path: Path

    @property
    def extra_config(self) -> dict:
        """
        Derived classes can override this to provide extra config to the
        create_app function.

        """
        return {}

    # If necessary, this function can be overridden to do something more
    # sophisticated to check the evaluation
    def check_evaluation(
        self,
        bundle: Bundle,
        evaluation: Evaluation,
        input_data: InputData,
    ) -> None:
        evaluation(bundle, input_data)

    @property
    def _input_data(self) -> InputData:
        return InputData(self.input_data_path)

    def setup(self) -> None:
        self.graph = create_evaluate_app(
            extra_loader=load_each(
                load_from_dict(
                    active_evaluation=self.evaluation_name,
                    root_input_artifact_path=self.root_input_artifact_path,
                ),
                load_from_dict(self.extra_config),
            ),
        )

    def test_evaluation(self) -> None:
        self.check_evaluation(
            self.graph.active_bundle,
            self.graph.active_evaluation,
            self._input_data,
        )



class TrainingInitializerRegistry:
    """
    The training initializer registry is a place to register functions that need to be
    called during training initialization.  A typical example of this is to
    seed random number generators.

    """
    def __init__(self, graph):
        self.graph = graph
        self.initializers = []

    def register(self, initializer):
        self.initializers.append(initializer)

    def init(self):
        for initializer in self.initializers:
            initializer.init()



from pkg_resources import iter_entry_points

from microcosm.object_graph import ObjectGraph

from microcosm_sagemaker.constants import (
    APP_HOOKS_GROUP,
    EVALUATE_APP_HOOK,
    SERVE_APP_HOOK,
    TRAIN_APP_HOOK,
)
from microcosm_sagemaker.exceptions import AppHookNotFoundError


def create_train_app(*args, **kwargs) -> ObjectGraph:
    return _create_app(TRAIN_APP_HOOK, args, kwargs)


def create_serve_app(*args, **kwargs) -> ObjectGraph:
    return _create_app(SERVE_APP_HOOK, args, kwargs)


def create_evaluate_app(*args, **kwargs) -> ObjectGraph:
    return _create_app(EVALUATE_APP_HOOK, args, kwargs)


def _create_app(name: str, args: tuple, kwargs: dict) -> ObjectGraph:
    try:
        entry_point = next(
            entry_point
            for entry_point in iter_entry_points(group=APP_HOOKS_GROUP)
            if entry_point.name == name
        )
    except StopIteration:
        raise AppHookNotFoundError(name)

    factory = entry_point.load()

    return factory(*args, **kwargs)



"""
Main training CLI

"""
import json

from click import (
    UNPROCESSED,
    File,
    argument,
    command,
    option,
)
from microcosm.loaders import load_from_dict
from microcosm.loaders.compose import merge
from microcosm.object_graph import ObjectGraph

from microcosm_sagemaker.app_hooks import create_train_app
from microcosm_sagemaker.artifact import RootOutputArtifact
from microcosm_sagemaker.commands.config import load_config_from_command_line_arguments
from microcosm_sagemaker.commands.evaluate import run_evaluate
from microcosm_sagemaker.commands.options import input_data_option, output_artifact_option
from microcosm_sagemaker.exceptions import raise_sagemaker_exception
from microcosm_sagemaker.input_data import InputData


@command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@option(
    "--configuration",
    type=File('r'),
    help="Manual import of configuration file, used for local testing",
)
@input_data_option()
@output_artifact_option()
@option(
    "--auto-evaluate/--no-auto-evaluate",
    default=True,
    help="Whether to automatically evaluate after the training has completed",
)
@option(
    "--testing/--no-testing",
    default=False,
    help="Whether the training is for testing purposes"
)
@argument('command_line_config', nargs=-1, type=UNPROCESSED)
def main(configuration, input_data, output_artifact, auto_evaluate, testing, command_line_config):
    try:

        extra_config_from_file = json.load(configuration) if configuration else dict()
        extra_config_from_command_line = load_config_from_command_line_arguments(command_line_config)
        extra_config = merge([extra_config_from_file, extra_config_from_command_line])

        graph = create_train_app(
            extra_loader=load_from_dict(extra_config),
            testing=testing,
        )

        run_train(
            graph=graph,
            input_data=input_data,
            root_output_artifact=output_artifact,
        )

        if auto_evaluate:
            run_evaluate(graph, input_data)

    except Exception as e:
        raise_sagemaker_exception(e)


def run_train(
    graph: ObjectGraph,
    input_data: InputData,
    root_output_artifact: RootOutputArtifact,
) -> None:
    root_output_artifact.init()

    graph.experiment_metrics.init()

    root_output_artifact.save_config(graph.config)

    graph.bundle_and_dependencies_trainer(
        bundle=graph.active_bundle,
        input_data=input_data,
        root_output_artifact=root_output_artifact,
    )



"""
Invocations controller.

"""
from microcosm_flask.conventions.saved_search import configure_saved_search
from microcosm_flask.namespaces import Namespace
from microcosm_flask.operations import Operation


def configure_invocations(graph, definition):
    """
    Define the invocations endpoint required by Sagemaker

    """
    # NB: The `SavedSearch` operation is used when we want to support a
    # `search` endpoint that expects a `POST` instead of a `GET`.  In our case,
    # we don't actually save the search in any way.
    graph.config.swagger_convention.operations.append("saved_search")
    ns = Namespace(
        subject="invocations",
        version="v1",
    )
    mappings = {
        Operation.SavedSearch: definition,
    }
    configure_saved_search(graph, ns, mappings)

    return ns



import json
from typing import (
    Callable,
    Generic,
    NamedTuple,
    TypeVar,
)

from hamcrest.core.base_matcher import BaseMatcher


T = TypeVar('T')


class ExtractorMatcherPair(NamedTuple, Generic[T]):
    """
    Contains a pair of extractor and matcher which is used during testing.  The
    extractor will be applied to the raw bytes of a file output during testing,
    as well as to the raw bytes of the gold version of the file.  The output of
    the extractor applied to the gold file will then be passed to
    `matcher_constructor`, which will create a matcher based on the gold file.

    """
    # NB: We cannot use a dataclass due to python/mypy#5485
    extractor: Callable[[bytes], T]
    matcher_constructor: Callable[[T], BaseMatcher]


def json_extractor(raw_bytes):
    return json.loads(raw_bytes.decode())



from microcosm_logging.decorators import logger


@logger
class ExperimentMetrics:
    """
    This registry is a place to register functions that log metrics.
    The `log_static` and `log_timeseries` methods are used to store single-value
    and multiple-value metrics, respectively.

    """
    def __init__(self, graph):
        self.graph = graph
        self.testing = graph.metadata.testing
        self.metric_observers = []

    def register(self, observer):
        if not self.testing:
            self.metric_observers.append(observer)

    def init(self):
        """
        Calls the init method on all of the registered metric observers.

        """
        for metric_observer in self.metric_observers:
            metric_observer.init()

    def log_timeseries(self, **kwargs):
        if not self.testing:
            for metric_observer in self.metric_observers:
                response = metric_observer.log_timeseries(**kwargs)
                if response:
                    self.logger.info(response)

    def log_static(self, **kwargs):
        if not self.testing:
            for metric_observer in self.metric_observers:
                response = metric_observer.log_static(**kwargs)
                if response:
                    self.logger.info(response)



"""
Upload pytest runs to wandb

Adapted from the pastebin code:
https://github.com/pytest-dev/pytest/blob/e40bf1d1da6771f951bd4b6126fc3cb107a7c9e7/src/_pytest/pastebin.py

"""
from json import load
from logging import info, warning
from tempfile import NamedTemporaryFile, TemporaryFile

import pytest  # noqa: 401
import wandb  # noqa: 401

from microcosm_sagemaker.constants import ARTIFACT_CONFIGURATION_PATH, SagemakerPath


@pytest.hookimpl(trylast=True)
def pytest_configure(config):
    terminal_reporter = config.pluginmanager.getplugin("terminalreporter")

    # If no terminal reporter plugin is present, nothing we can do here;
    # this can happen when this function executes in a slave node
    # when using pytest-xdist, for example
    if terminal_reporter is None:
        return

    # Monkey-patch the write function
    config._pytestsessionfile = TemporaryFile("w+")
    oldwrite = terminal_reporter._tw.write

    def tee_write(s, **kwargs):
        oldwrite(s, **kwargs)
        config._pytestsessionfile.write(str(s))

    terminal_reporter._tw.write = tee_write


def pytest_unconfigure(config):
    # Only write the data upon test close if we have access to the written file
    if not hasattr(config, "_pytestsessionfile"):
        return

    # We already got the contents in `pytest_terminal_summary`,
    # so just close and delete file
    config._pytestsessionfile.close()
    del config._pytestsessionfile

    # Undo our patching in the terminal reporter
    terminal_reporter = config.pluginmanager.getplugin("terminalreporter")
    del terminal_reporter._tw.__dict__["write"]


def pytest_terminal_summary(terminalreporter, config):
    """
    Given a terminal reporter, call our custom writer callback to write to wandb

    NOTE: The lack of snake case of `terminalreporter` is important here because it is
    dynamically passed based on the name of the plugin - which is sans snake case.
    """
    # add captured stdout first
    reports = terminalreporter.getreports("")
    content = "".join(report.capstdout for report in reports)

    # add summary report afterward
    config._pytestsessionfile.seek(0)
    sessionlog = config._pytestsessionfile.read()
    content += "\n" + sessionlog
    create_new_file(content)


def create_new_file(contents):
    """
    Uploads the provided contents to the wandb
    session that is linked to this artifact.
    """

    # Access our artifact's configuration file
    configuration_path = SagemakerPath.MODEL / ARTIFACT_CONFIGURATION_PATH

    # If we are not operating in an environment that has access to an artifact file
    # No-op out of the upload
    if not configuration_path.exists():
        return

    with open(configuration_path) as file:
        configuration = load(file)
        run_path = configuration.get("wandb", {}).get("run_path", {})

    if not run_path:
        warning("Unable to upload logging file to `wandb`: no runpath found")
        return

    # Load the wandb run that is linked to this artifact
    api = wandb.Api()
    run = api.run(run_path)
    info(f"Writing results to wandb run: {run_path}")

    with NamedTemporaryFile(prefix="artifact-test-", suffix=".log", mode="w+") as file:
        file.write(contents)
        file.seek(0)
        run.upload_file(file.name, root="/")



import logging
import sys
from traceback import print_exception
from typing import Sequence

import click
from click.testing import CliRunner
from hamcrest import assert_that, equal_to, is_


class CliTestCase:
    def setup(self) -> None:
        self.runner = CliRunner()

    def run_and_check(
        self,
        command_name: str,
        command: click.Command,
        args: Sequence[str],
    ) -> None:
        logging.info(f"Running command: {command_name} {' '.join(args)}")

        result = self.runner.invoke(command, args)

        if result.exit_code != 0:
            sys.stdout.write(result.output)
            if result.exc_info is not None:
                print_exception(*result.exc_info)

        assert_that(result.exit_code, is_(equal_to(0)))



"""
SageMaker-specific exception handling.

"""
from sys import exit, stderr
from traceback import format_exc

from microcosm_sagemaker.constants import APP_HOOKS_GROUP, SagemakerPath


class DependencyCycleError(Exception):
    def __init__(self, component):
        super().__init__(f"Dependency cycle involving '{component}'.")


class AppHookNotFoundError(Exception):
    def __init__(self, hook_name):
        super().__init__(
            f"App hook '{hook_name}' undefined.  Please define '{hook_name}' "
            f"in {APP_HOOKS_GROUP}"
        )


def raise_sagemaker_exception(exception):
    traceback = format_exc()
    description = "Exception during training: " + str(exception) + "\n" + traceback + "\n"

    if not SagemakerPath.OUTPUT.exists():
        raise exception

    # Record the failure in a format that is saved along with the training run
    with open(SagemakerPath.OUTPUT / "failure", "w") as error_log:
        error_log.write(description)

    # Also display the status in the job logs
    stderr.write(description)

    # Mark training job as failed
    exit(255)



from typing import Any, Callable

from microcosm.object_graph import ObjectGraph
from microcosm_logging.timing import elapsed_time


def training_initializer():
    """
    Register a microcosm component as a training initializer, so that its init
    method will automatically be called.  This function is designed to be used
    as a decorator on a factory.

    """
    def decorator(func: Callable[[ObjectGraph], Any]):
        def factory(graph):
            component = func(graph)
            graph.training_initializers.register(component)
            return component
        return factory
    return decorator


def metrics_observer():
    """
    Register a microcosm component as an experiment metric tracker,
    for it to get any metrics being logged.

    """
    def decorator(func: Callable[[ObjectGraph], Any]):
        def factory(graph):
            component = func(graph)
            graph.experiment_metrics.register(component)
            return component
        return factory
    return decorator


def _method_with_logging(original_method):
    def new_method(self, *args, **kwargs):
        self.logger.info(f"Started method `{original_method.__name__}`.")
        timing = {}
        with elapsed_time(timing):
            original_method(self, *args, **kwargs)
        self.logger.info(
            f"Completed method `{original_method.__name__}`"
            f"after {timing['elapsed_time']/1000:.3g} seconds."
        )
    return new_method


def log_bundle_methods(cls):

    _init = cls.__init__

    def __init__(self, graph: ObjectGraph, **kwargs) -> None:
        _init(self, graph, **kwargs)
        self._graph = graph

    cls.__init__ = __init__
    cls.fit = _method_with_logging(cls.fit)
    cls.load = _method_with_logging(cls.load)
    cls.save = _method_with_logging(cls.save)

    return cls



import tempfile
from pathlib import Path
from typing import List, Mapping, Optional

from microcosm.loaders import load_each, load_from_dict

from microcosm_sagemaker.app_hooks import create_evaluate_app
from microcosm_sagemaker.bundle import Bundle
from microcosm_sagemaker.commands import train
from microcosm_sagemaker.testing.bundle_prediction_check import BundlePredictionCheck
from microcosm_sagemaker.testing.bytes_extractor import ExtractorMatcherPair
from microcosm_sagemaker.testing.cli_test_case import CliTestCase
from microcosm_sagemaker.testing.directory_comparison import directory_comparison


class TrainCliTestCase(CliTestCase):
    """
    Helper base class for writing tests of the train cli.

    """
    input_data_path: Path
    gold_output_artifact_path: Optional[Path] = None
    output_artifact_matchers: Mapping[Path, ExtractorMatcherPair]
    bundle_prediction_checks: Optional[List[BundlePredictionCheck]] = None

    @property
    def extra_evaluate_config(self) -> dict:
        """
        Derived classes can override this to provide extra config to the
        evalute create_app function when evaluating the active bundle.

        """
        return {}

    def test_train(self) -> None:
        """
        Runs the `train` command on the given `input_data_path` and then
        recursively checks the contents of the output artifact against the
        expected contents in `gold_output_artifact_path`.  It is also possible
        to leave certain files out of the gold dir, and instead specify a
        matcher that should be used for the contents of the given file instead.

        """
        with tempfile.TemporaryDirectory() as output_artifact_path:
            self.run_and_check(
                command_name="train",
                command=train.main,
                args=[
                    "--input-data",
                    str(self.input_data_path),
                    "--output-artifact",
                    output_artifact_path,
                    "--testing",
                ],
            )

            self.check_output_artifact(Path(output_artifact_path))

            self.graph = create_evaluate_app(
                extra_loader=load_each(
                    load_from_dict(
                        root_input_artifact_path=output_artifact_path,
                    ),
                    load_from_dict(self.extra_evaluate_config),
                ),
                testing=True,
            )

            self.check_bundle(self.graph.active_bundle)

    def check_output_artifact(self, output_artifact_path: Path):
        """
        Checks the training output artifact.  Can be overridden to perform
        customized checks.

        """
        if self.gold_output_artifact_path:
            directory_comparison(
                gold_dir=self.gold_output_artifact_path,
                actual_dir=output_artifact_path,
                matchers=self.output_artifact_matchers,
            )

    def check_bundle(self, bundle: Bundle):
        """
        Checks the active bundle after loading trained artifact.  Derived
        classes can override this to perform custom bundle checks.

        """
        if self.bundle_prediction_checks:
            for bundle_prediction_check in self.bundle_prediction_checks:
                bundle_prediction_check.check_bundle(bundle)



import tempfile
from abc import ABC
from pathlib import Path
from typing import List, Mapping, Optional

from microcosm.loaders import load_each, load_from_dict

from microcosm_sagemaker.app_hooks import create_evaluate_app, create_train_app
from microcosm_sagemaker.artifact import (
    BundleInputArtifact,
    BundleOutputArtifact,
    RootInputArtifact,
)
from microcosm_sagemaker.bundle import Bundle
from microcosm_sagemaker.input_data import InputData
from microcosm_sagemaker.testing.bundle_prediction_check import BundlePredictionCheck
from microcosm_sagemaker.testing.bytes_extractor import ExtractorMatcherPair
from microcosm_sagemaker.testing.directory_comparison import directory_comparison


class BundlePredictionChecker:
    bundle_prediction_checks: List[BundlePredictionCheck]

    # If necessary, this function can be overridden to do something more
    # sophisticated to check the bundle prediction
    def check_bundle_prediction(self, bundle: Bundle) -> None:
        for bundle_prediction_check in self.bundle_prediction_checks:
            bundle_prediction_check.check_bundle(bundle)


class BundleTestCase(ABC):
    # These should be defined in actual test case derived class
    bundle_name: str
    root_input_artifact_path: Path

    @property
    def extra_config(self) -> dict:
        """
        Derived classes can override this to provide extra config to the
        various create_app functions.

        """
        return {}

    @property
    def _root_input_artifact(self) -> RootInputArtifact:
        return RootInputArtifact(self.root_input_artifact_path)


class BundleFitTestCase(BundleTestCase, BundlePredictionChecker):
    input_data_path: Path

    @property
    def _input_data(self) -> InputData:
        return InputData(self.input_data_path)

    def setup(self) -> None:
        self.graph = create_train_app(
            extra_loader=load_each(
                load_from_dict(
                    active_bundle=self.bundle_name,
                ),
                load_from_dict(self.extra_config),
            ),
            testing=True,
        )

        self.graph.bundle_and_dependencies_loader(
            bundle=self.graph.active_bundle,
            root_input_artifact=self._root_input_artifact,
            dependencies_only=True,
        )

        self.graph.experiment_metrics.init()

    def test_fit(self) -> None:
        self.graph.active_bundle.fit(self._input_data)
        self.check_bundle_prediction(self.graph.active_bundle)


class BundleSaveTestCase(BundleTestCase):
    gold_bundle_output_artifact_path: Path
    output_artifact_matchers: Optional[Mapping[Path, ExtractorMatcherPair]] = None
    ignore_file_contents: bool = False

    @property
    def _gold_bundle_output_artifact(self) -> BundleOutputArtifact:
        return BundleOutputArtifact(self.gold_bundle_output_artifact_path)

    def setup(self) -> None:
        self.graph = create_evaluate_app(
            extra_loader=load_each(
                load_from_dict(
                    active_bundle=self.bundle_name,
                    root_input_artifact_path=self.root_input_artifact_path,
                ),
                load_from_dict(self.extra_config),
            ),
            testing=True,
        )

        self.temporary_directory = tempfile.TemporaryDirectory()
        self.bundle_output_artifact = BundleOutputArtifact(self.temporary_directory.name)

    def teardown(self) -> None:
        self.temporary_directory.cleanup()

    def test_save(self) -> None:
        self.graph.active_bundle.save(self.bundle_output_artifact)

        directory_comparison(
            gold_dir=self._gold_bundle_output_artifact.path,
            actual_dir=self.bundle_output_artifact.path,
            matchers=self.output_artifact_matchers,
            ignore_file_contents=self.ignore_file_contents,
        )


class BundleLoadTestCase(BundleTestCase, BundlePredictionChecker):
    def setup(self) -> None:
        self.graph = create_evaluate_app(
            extra_loader=load_each(
                load_from_dict(
                    active_bundle=self.bundle_name,
                    root_input_artifact_path=self.root_input_artifact_path,
                    load_active_bundle_and_dependencies=dict(
                        perform_load=False,
                    ),
                ),
                load_from_dict(self.extra_config),
            ),
            testing=True,
        )

        self.graph.bundle_and_dependencies_loader(
            bundle=self.graph.active_bundle,
            root_input_artifact=self._root_input_artifact,
            dependencies_only=True,
        )

    def test_load(self) -> None:
        self.graph.active_bundle.load(
            self._root_input_artifact / self.bundle_name,
        )

        self.check_bundle_prediction(self.graph.active_bundle)


class BundleFitSaveLoadTestCase(BundleTestCase, BundlePredictionChecker):
    """
    One combined test of fit, save and load.  First fits and saves bundle, then
    optionally tests artifact against a gold directory, then loads the trained
    bundle and runs behaviour tests.

    """
    input_data_path: Path

    # These are optional; only required for doing gold-based testing
    gold_bundle_output_artifact_path: Optional[Path] = None
    output_artifact_matchers: Optional[Mapping[Path, ExtractorMatcherPair]] = None
    ignore_file_contents: bool = False

    @property
    def _gold_bundle_output_artifact(self) -> BundleOutputArtifact:
        assert self.gold_bundle_output_artifact_path is not None

        return BundleOutputArtifact(self.gold_bundle_output_artifact_path)

    @property
    def _input_data(self) -> InputData:
        return InputData(self.input_data_path)

    def setup(self) -> None:
        self.setup_train_graph()
        self.setup_evaluate_graph()

        self.temporary_directory = tempfile.TemporaryDirectory()

    def teardown(self) -> None:
        self.temporary_directory.cleanup()

    def setup_train_graph(self) -> None:
        self.train_graph = create_train_app(
            extra_loader=load_each(
                load_from_dict(
                    active_bundle=self.bundle_name,
                ),
                load_from_dict(self.extra_config),
            ),
            testing=True,
        )

        self.train_graph.bundle_and_dependencies_loader(
            bundle=self.train_graph.active_bundle,
            root_input_artifact=self._root_input_artifact,
            dependencies_only=True,
        )

        self.train_graph.experiment_metrics.init()

    def setup_evaluate_graph(self):
        self.evaluate_graph = create_evaluate_app(
            extra_loader=load_each(
                load_from_dict(
                    active_bundle=self.bundle_name,
                    root_input_artifact_path=self.root_input_artifact_path,
                    load_active_bundle_and_dependencies=dict(
                        perform_load=False,
                    ),
                ),
                load_from_dict(self.extra_config),
            ),
            testing=True,
        )

        self.evaluate_graph.bundle_and_dependencies_loader(
            bundle=self.evaluate_graph.active_bundle,
            root_input_artifact=self._root_input_artifact,
            dependencies_only=True,
        )

    def test_fit_save_load(self) -> None:
        self.graph = self.train_graph
        self.graph.active_bundle.fit(self._input_data)
        self.graph.active_bundle.save(
            BundleOutputArtifact(self.temporary_directory.name),
        )
        self.check_bundle_prediction(self.graph.active_bundle)

        if self.gold_bundle_output_artifact_path:
            directory_comparison(
                gold_dir=self._gold_bundle_output_artifact.path,
                actual_dir=Path(self.temporary_directory.name),
                matchers=self.output_artifact_matchers,
                ignore_file_contents=self.ignore_file_contents,
            )

        self.graph = self.evaluate_graph
        self.graph.active_bundle.load(
            BundleInputArtifact(self.temporary_directory.name),
        )
        self.check_bundle_prediction(self.graph.active_bundle)



from random import seed

from microcosm.api import defaults
from microcosm.config.validation import typed

from microcosm_sagemaker.decorators import training_initializer


@defaults(
    seed=typed(int, 42),
)
@training_initializer()
class Random:
    def __init__(self, graph):
        self.seed = graph.config.random.seed

    def init(self):
        # Seed python random
        seed(self.seed)

        # Seed numpy if installed
        try:
            import numpy as np
            np.random.seed(self.seed)
        except ImportError:
            pass

        # Seed torch if installed
        try:
            from torch import manual_seed
            manual_seed(self.seed)
        except ImportError:
            pass

        # Seed tensorflow if installed
        try:
            from tf.random import set_random_seed
            set_random_seed(self.seed)
        except ImportError:
            pass



from abc import ABC, abstractmethod
from typing import Callable, List

from microcosm_sagemaker.artifact import BundleInputArtifact, BundleOutputArtifact
from microcosm_sagemaker.input_data import InputData


class Bundle(ABC):
    """
    Abstract base class for all bundles.

    """
    predict: Callable

    @property
    @abstractmethod
    def dependencies(self) -> List["Bundle"]:
        """
        List of bundles upon which this bundle depends.  Whenever the `fit`,
        `save` or `load` methods are called on this bundle, it is guaranteed
        that the corresponding methods will have first been called all all
        `dependency` bundles.

        """
        ...

    @abstractmethod
    def fit(self, input_data: InputData) -> None:
        """
        Perform training

        """
        ...

    @abstractmethod
    def save(self, output_artifact: BundleOutputArtifact) -> None:
        """
        Save the trained model

        """
        ...

    @abstractmethod
    def load(self, input_artifact: BundleInputArtifact) -> None:
        """
        Load the trained model

        """
        ...



"""
Consumer factories.

"""
from microcosm.api import defaults
from microcosm.object_graph import ObjectGraph

from microcosm_sagemaker.artifact import RootInputArtifact


def configure_active_bundle(graph):
    if not getattr(graph.config, "active_bundle", ""):
        return None
    return getattr(graph, graph.config.active_bundle)


def configure_active_evaluation(graph):
    if not getattr(graph.config, "active_evaluation", ""):
        return None
    return getattr(graph, graph.config.active_evaluation)


@defaults(
    perform_load=True,
)
def load_active_bundle_and_dependencies(graph: ObjectGraph):
    """
    Loads the active bundle and its dependencies immediately upon instantation.

    """
    if not graph.config.load_active_bundle_and_dependencies.perform_load:
        return

    root_input_artifact = RootInputArtifact(graph.config.root_input_artifact_path)

    graph.bundle_and_dependencies_loader(
        bundle=graph.active_bundle,
        root_input_artifact=root_input_artifact,
    )


def configure_sagemaker(graph):
    """
    Instantiates all the necessary sagemaker factories.

    """
    graph.use(
        "active_bundle",
        "active_evaluation",
        "bundle_and_dependencies_loader",
        "bundle_and_dependencies_trainer",
        "random",
        "training_initializers",
        "experiment_metrics",
    )



from abc import ABC, abstractmethod
from typing import Callable

from microcosm.object_graph import ObjectGraph

from microcosm_sagemaker.bundle import Bundle
from microcosm_sagemaker.dependency_traverser import traverse_component_and_dependencies


class BundleOrchestrator(ABC):
    def __init__(self, graph: ObjectGraph):
        pass

    @abstractmethod
    def __call__(
        self,
        bundle: Bundle,
        bundle_handler: Callable[[Bundle], None],
        dependencies_only: bool = False,
    ):
        """
        Given a `bundle`, call `bundle_handler` on `bundle` and all its
        transitive dependencies, ensuring that `bundle_handler` has been called
        on all dependencies of any bundle before it is called on the bundle
        itself.

        If `dependencies_only` is `True`, do not call `bundle_handler` on the
        given `bundle` itself, but rather only its dependencies.

        """
        ...


class SingleThreadedBundleOrchestrator(BundleOrchestrator):
    """
    Performs topological sort on dependencies of `bundle` and calls
    `bundle_handler` on each bundle one at a time.

    """
    def __call__(
        self,
        bundle: Bundle,
        bundle_handler: Callable[[Bundle], None],
        dependencies_only: bool = False,
    ):
        for bundle_to_handle in traverse_component_and_dependencies(bundle):
            if dependencies_only and bundle_to_handle == bundle:
                continue

            bundle_handler(bundle_to_handle)



import json
from abc import ABC
from pathlib import Path
from typing import Union

from microcosm.config.model import Configuration
from microcosm.metadata import Metadata

from microcosm_sagemaker.constants import ARTIFACT_CONFIGURATION_PATH


class OutputArtifact(ABC):
    def __init__(self, path: Union[str, Path]):
        self.path = Path(path)

    def init(self) -> None:
        self.path.mkdir(parents=True, exist_ok=True)

    def __repr__(self) -> str:
        return f'{type(self).__name__}("{str(self.path)}")'


class BundleOutputArtifact(OutputArtifact):
    pass


class RootOutputArtifact(OutputArtifact):
    def __init__(self, path: Union[str, Path]):
        super().__init__(path)

    def __truediv__(self, name: str):
        return BundleOutputArtifact(self.path / name)

    def save_config(self, config: Configuration) -> None:
        config_path = self.path / ARTIFACT_CONFIGURATION_PATH

        with open(config_path, "w") as config_file:
            json.dump(config, config_file)


class InputArtifact(ABC):
    def __init__(self, path: Union[str, Path]):
        self.path = Path(path)

    def __repr__(self) -> str:
        return f'{type(self).__name__}("{str(self.path)}")'


class BundleInputArtifact(InputArtifact):
    pass


class RootInputArtifact(InputArtifact):
    def __init__(self, path: Union[str, Path]):
        super().__init__(path)

    def __truediv__(self, name: str):
        return BundleInputArtifact(self.path / name)

    def load_config(self, metadata: Metadata) -> Configuration:
        """
        When we train a model, we freeze all of the current graph variables and store it alongside
        the artifact. Whenever we boot up the model again, we want to hydrate this from disk.

        Note that this function is designed to be used as a microcosm loader,
        hence the unused `metadata` parameter.

        """
        config_path = self.path / ARTIFACT_CONFIGURATION_PATH

        with open(config_path) as config_file:
            return Configuration(json.load(config_file))



from abc import ABC, abstractmethod

from microcosm_sagemaker.bundle import Bundle
from microcosm_sagemaker.input_data import InputData


class Evaluation(ABC):
    @abstractmethod
    def __call__(self, bundle: Bundle, input_data: InputData):
        pass



from functools import partial

from microcosm.api import defaults
from microcosm.object_graph import ObjectGraph

from microcosm_sagemaker.artifact import RootInputArtifact, RootOutputArtifact
from microcosm_sagemaker.bundle import Bundle
from microcosm_sagemaker.dependency_traverser import traverse_component_and_dependencies
from microcosm_sagemaker.input_data import InputData


@defaults(
    bundle_orchestrator="single_threaded_bundle_orchestrator",
)
class BundleAndDependenciesLoader:
    def __init__(self, graph: ObjectGraph):
        self.bundle_orchestrator = getattr(
            graph,
            graph.config.bundle_and_dependencies_loader.bundle_orchestrator,
        )
        self.graph = graph

    def __call__(
        self,
        bundle: Bundle,
        root_input_artifact: RootInputArtifact,
        dependencies_only: bool = False,
    ):
        def load(bundle):
            name = _get_component_name(self.graph, bundle)
            bundle.load(root_input_artifact / name)

        self.bundle_orchestrator(
            bundle=bundle,
            bundle_handler=load,
            dependencies_only=dependencies_only,
        )


def fit_and_save_bundle(
    graph: ObjectGraph,
    input_data: InputData,
    root_output_artifact: RootOutputArtifact,
    bundle: Bundle,
) -> None:
    bundle_name = _get_component_name(graph, bundle)

    nested_output_artifact = root_output_artifact / bundle_name
    nested_output_artifact.init()

    # We call the training_initializers.init() before any bundle.fit()
    # That makes each bundle's training reproducible, regardless of its order in the training queue.
    graph.training_initializers.init()

    bundle.fit(input_data)
    bundle.save(nested_output_artifact)


@defaults(
    bundle_orchestrator="single_threaded_bundle_orchestrator",
)
class BundleAndDependenciesTrainer:
    def __init__(self, graph: ObjectGraph):
        self.bundle_orchestrator = getattr(
            graph,
            graph.config.bundle_and_dependencies_trainer.bundle_orchestrator,
        )
        self.graph = graph

    def __call__(
        self,
        bundle: Bundle,
        input_data: InputData,
        root_output_artifact: RootOutputArtifact,
        dependencies_only: bool = False,
    ):
        train = partial(
            fit_and_save_bundle,
            self.graph,
            input_data,
            root_output_artifact,
        )

        self.bundle_orchestrator(
            bundle=bundle,
            bundle_handler=train,
            dependencies_only=dependencies_only,
        )


class BundleAndDependenciesConfigExtractor:
    def __init__(self, graph):
        self.config = graph.config
        self.graph = graph

    def __call__(self, bundle: Bundle):
        """
        Returns the config from the `bundle`, as well as all of its dependents.

        """
        config = {}

        for bundle_to_handle in traverse_component_and_dependencies(bundle):
            bundle_name = _get_component_name(self.graph, bundle_to_handle)
            bundle_config = getattr(self.config, bundle_name, {})
            config.update(
                {bundle_name: bundle_config}
            )

        return config


def _get_component_name(graph, component):
    return next(
        key
        for key, possible_component in graph.items()
        if possible_component == component
    )



"""
Example CRUD routes tests.

Tests are sunny day cases under the assumption that framework conventions
handle most error conditions.

"""
from pathlib import Path

from microcosm.loaders import load_from_dict

from microcosm_sagemaker.app_hooks import create_serve_app


class RouteTestCase:
    """
    Helper base class for writing tests of a route.

    """
    root_input_artifact_path: Path

    def setup(self) -> None:
        self.graph = create_serve_app(
            testing=True,
            extra_loader=load_from_dict(
                root_input_artifact_path=self.root_input_artifact_path,
            ),
        )

        self.client = self.graph.flask.test_client()



from abc import ABC, abstractmethod
from inspect import signature
from typing import (
    Callable,
    Iterable,
    Mapping,
    Optional,
)


class PipelineHarness(ABC):
    """
    Base class to support feature tests, by specifying the steps/pipes
    of a pipeline

    """

    @property
    @abstractmethod
    def _steps(self) -> Iterable[Callable[..., Optional[Mapping]]]:
        """
        Define the steps for this testing pipeline.

        """
        pass

    def test_steps(self):
        store = {}
        for fn in self._steps:
            fn_parameters = signature(fn).parameters

            # Pass through only the parameters that the function requires
            current_output = fn(**{key: store[key] for key in fn_parameters})

            # update store if any outputs are returned
            store.update(current_output or {})



"""
Loaders to inject SM parameters as microcosm configurations.

"""
import json

from boto3 import client
from microcosm.config.model import Configuration
from microcosm.loaders import load_each
from microcosm.loaders.compose import two_stage_loader
from microcosm.loaders.keys import expand_config
from microcosm.metadata import Metadata
from microcosm.typing import Loader

from microcosm_sagemaker.artifact import RootInputArtifact
from microcosm_sagemaker.commands.config import load_default_microcosm_runserver_config
from microcosm_sagemaker.constants import SagemakerPath
from microcosm_sagemaker.s3 import S3Object


def hyperparameter_loader(metadata: Metadata) -> Configuration:
    """
    Sagemaker only supports single-layer hyperparameters, so we use double underscores
    (__) to signify the delineation between nested dictionaries.  This mirrors the
    formatting of our ENV variables.  Note that these values are all strings by convention,
    so any end applications should.

    This configuration helper parses these into the underlying dictionary structure.

    """
    try:
        with open(SagemakerPath.HYPERPARAMETERS) as raw_file:
            return expand_config(
                json.load(raw_file),
                separator="__",
                skip_to=0,
            )

    except FileNotFoundError:
        return Configuration()


def s3_loader(metadata: Metadata, configuration: Configuration) -> Configuration:
    """
    Opinionated loader that:
    1. Reads all of the hyperparameters passed through by SageMaker
    2. Uses a special `base_configuration` key to read the given configuration from S3

    """
    base_configuration_url = configuration.get("base_configuration", None)

    if base_configuration_url:
        s3 = client("s3")
        s3_object = S3Object.from_url(base_configuration_url)

        object = s3.get_object(Bucket=s3_object.bucket, Key=s3_object.key)
        return Configuration(json.loads(object["Body"].read()))

    return Configuration()


def root_input_artifact_config_loader(metadata: Metadata, config: Configuration) -> Configuration:
    root_input_artifact = RootInputArtifact(config.root_input_artifact_path)

    return root_input_artifact.load_config(metadata)


def train_conventions_loader(initial_loader: Loader) -> Loader:
    return two_stage_loader(
        primary_loader=load_each(
            initial_loader,
            hyperparameter_loader,
        ),
        secondary_loader=s3_loader,
    )


def serve_conventions_loader(initial_loader: Loader) -> Loader:
    return two_stage_loader(
        primary_loader=load_each(
            load_default_microcosm_runserver_config,
            initial_loader,
        ),
        secondary_loader=root_input_artifact_config_loader,
    )


def evaluate_conventions_loader(initial_loader: Loader) -> Loader:
    return two_stage_loader(
        primary_loader=initial_loader,
        secondary_loader=root_input_artifact_config_loader,
    )



#!/usr/bin/env python
from setuptools import find_packages, setup


project = "microcosm-sagemaker"
version = "0.9.4"

setup(
    name=project,
    version=version,
    description="Opinionated machine learning organization and configuration",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Globality Engineering",
    author_email="engineering@globality.com",
    url="https://github.com/globality-corp/microcosm-sagemaker",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.6",
    keywords="microcosm",
    install_requires=[
        "boto3>=1.9.90",
        "awscli>=1.16.200",
        "click>=7.0",
        "microcosm>=2.12.0",
        "microcosm-flask[metrics]>=2.8.0",
        # See: https://github.com/boto/botocore/pull/1910
        "python-dateutil<3.0.0",
        "joblib>=0.15",
    ],
    setup_requires=[
        "nose>=1.3.6",
    ],
    dependency_links=[
    ],
    entry_points={
        "console_scripts": [
            "train = microcosm_sagemaker.commands.train:main",
            "evaluate = microcosm_sagemaker.commands.evaluate:main",
            "runserver = microcosm_sagemaker.commands.runserver:main",
        ],
        "microcosm.factories": [
            "active_bundle = microcosm_sagemaker.factories:configure_active_bundle",
            "active_evaluation = microcosm_sagemaker.factories:configure_active_evaluation",
            (
                "bundle_and_dependencies_loader = "
                "microcosm_sagemaker.bundle_traversal:BundleAndDependenciesLoader"
            ),
            (
                "bundle_and_dependencies_trainer = "
                "microcosm_sagemaker.bundle_traversal:BundleAndDependenciesTrainer"
            ),
            (
                "bundle_and_dependencies_config_extractor = "
                "microcosm_sagemaker.bundle_traversal:BundleAndDependenciesConfigExtractor"
            ),
            (
                "load_active_bundle_and_dependencies = "
                "microcosm_sagemaker.factories:load_active_bundle_and_dependencies"
            ),
            "random = microcosm_sagemaker.random:Random",
            "sagemaker = microcosm_sagemaker.factories:configure_sagemaker",
            (
                "single_threaded_bundle_orchestrator = "
                "microcosm_sagemaker.bundle_orchestrator:SingleThreadedBundleOrchestrator"
            ),
            "training_initializers = microcosm_sagemaker.training_initializer_registry:TrainingInitializerRegistry",
            "experiment_metrics = microcosm_sagemaker.metrics.experiment_metrics:ExperimentMetrics",
            "wandb = microcosm_sagemaker.metrics.wandb.store:WeightsAndBiases",
        ],
        'pytest11': ['pytest_sagemaker = microcosm_sagemaker.testing.pytest_sagemaker'],
    },
    extras_require={
        "torch": [
            "torch>=1.1.0",
        ],
        "tensorflow": [
            "tensorflow>=1.14.0",
        ],
        "test": [
            "PyHamcrest>=1.9.0",
            "coverage>=4.0.3",
            "parameterized>=0.7.0",
            "torch>=1.1.0",
            "wandb>=0.10",
        ],
        "profiling": "pyinstrument>=3.0",
        "wandb": "wandb>=0.10",
    },
)



from contextlib import contextmanager
from typing import Any

from microcosm_sagemaker.exceptions import DependencyCycleError


def traverse_component_and_dependencies(component):
    """
    Given a component in a dependency graph, traverses the graph in topological
    order, ie such that all dependencies of a component will be yielded before
    the component itself.

    Each component is expected to have a `dependencies` property that returns
    an iterable of dependency components.  Each component is also expected to
    be hashable.

    """
    yield from _traverse_helper(
        seen=set(),
        reserved=set(),
        component=component,
    )


def _traverse_helper(
    seen: set,
    reserved: set,
    component: Any,
):
    """
    Helper to recursively yield a component and its dependencies.

    `seen` contains a list of all components that have already been visited.
    Its purpose is to prevent us from yielding the same component twice.

    `reserved` is designed to help detect cycles.  We add a component to
    `reserved` just before we visit its dependencies, in order to check that we
    don't see it as one of its transitive dependencies.  We remove it after we
    are done visiting its dependencies, because it is ok if we see the
    component again.  In that case, it will still be in `seen`, so we won't
    visit it again, but it won't be an error, as it would be if the component
    were in `reserved`.

    """
    if component in seen:
        return

    with _reserve(reserved, component):
        for dependency in component.dependencies:
            yield from _traverse_helper(
                seen=seen,
                reserved=reserved,
                component=dependency,
            )

    yield component
    seen.add(component)


@contextmanager
def _reserve(reserved: set, component: Any):
    """
    Adds component to a list of reserved components, erroring if the component
    has already been reserved, then remove component from list of reserved
    components.

    """
    if component in reserved:
        raise DependencyCycleError(component)

    reserved.add(component)
    try:
        yield
    finally:
        reserved.remove(component)



from pathlib import Path
from typing import Mapping, Optional

from hamcrest import (
    assert_that,
    contains,
    equal_to,
    is_,
)

from microcosm_sagemaker.testing.bytes_extractor import ExtractorMatcherPair


def _identity(x):
    return x


def _is_hidden(path: Path) -> bool:
    return path.name.startswith(".")


def directory_comparison(
    gold_dir: Path,
    actual_dir: Path,
    matchers: Optional[Mapping[Path, ExtractorMatcherPair]] = None,
    ignore_hidden: bool = True,
    ignore_file_contents: bool = False,
):
    """
    Recursively checks the contents of `actual_dir` against the expected
    contents in `gold_dir`.  It is also possible to leave certain files out of
    the gold dir, and instead specify an (extractor, matcher) pair that should
    be used to extract and match the contents of the given file instead.

    By default, this function ignores hidden files.  This functionality is
    useful when you expect an empty directory, because git won't allow checking
    in an empty directory.  In this situation you can add an empty `.keep` file
    to the directory to make sure it is checked in.

    """
    if matchers is None:
        matchers = dict()

    assert_that(gold_dir.exists(), is_(True))
    assert_that(actual_dir.exists(), is_(True))

    actual_paths = sorted([
        subpath.relative_to(actual_dir)
        for subpath in actual_dir.glob('**/*')
        if not (ignore_hidden and _is_hidden(subpath))  # exclude hidden files if ignore_hidden is True
    ])
    gold_paths = sorted([
        subpath.relative_to(gold_dir)
        for subpath in gold_dir.glob('**/*')
        if not (ignore_hidden and _is_hidden(subpath))  # exclude hidden files if ignore_hidden is True
    ])

    assert_that(actual_paths, contains(*gold_paths))

    for path in gold_paths:
        gold_path = gold_dir / path
        actual_path = actual_dir / path

        if gold_path.is_dir():
            assert_that(actual_path.is_dir(), is_(True))
        else:
            assert_that(actual_path.is_dir(), is_(False))

            if not ignore_file_contents:
                if path in matchers:
                    extractor, matcher_constructor = matchers[path]
                else:
                    extractor, matcher_constructor = ExtractorMatcherPair(
                        _identity,
                        lambda x: is_(equal_to(x)),
                    )

                assert_that(
                    extractor(actual_path.read_bytes()),
                    matcher_constructor(
                        extractor(gold_path.read_bytes()),
                    ),
                    path,
                )



"""
Example CRUD routes tests.

Tests are sunny day cases under the assumption that framework conventions
handle most error conditions.

"""
from typing import Optional

from hamcrest import (
    assert_that,
    equal_to,
    has_entries,
    is_,
)
from hamcrest.core.base_matcher import BaseMatcher

from microcosm_sagemaker.testing.route import RouteTestCase


class InvocationsRouteTestCase(RouteTestCase):
    """
    Helper base class for writing tests of the invocations route.

    """

    request_json: dict
    response_items_matcher: Optional[BaseMatcher] = None
    expected_status_code: int = 200

    def test_search(self) -> None:
        """
        Invokes the invocations endpoint with `request_json`, and checks the
        `items` entry of the response against `response_items_matcher`.

        """
        uri = "/api/v1/invocations"

        response = self.client.post(
            uri,
            json=self.request_json,
        )

        assert_that(response.status_code, is_(equal_to(self.expected_status_code)))

        if self.response_items_matcher:
            assert_that(
                response.json,
                has_entries(
                    items=self.response_items_matcher,
                ),
            )



from typing import List

import joblib

from microcosm_sagemaker.artifact import BundleInputArtifact, BundleOutputArtifact
from microcosm_sagemaker.bundle import Bundle


class PicklingBundle(Bundle):
    """
    Automatically defines save and load to pickle and load a list of attrs
    specified in `pickle_attrs`

    """
    pickle_attrs: List[str]

    def save(self, output_artifact: BundleOutputArtifact) -> None:
        """
        Save the trained model

        """
        for pickle_attr in self.pickle_attrs:
            joblib.dump(
                getattr(self, pickle_attr),
                output_artifact.path / f"{pickle_attr}.pkl",
            )

    def load(self, input_artifact: BundleInputArtifact) -> None:
        """
        Load the trained model

        """
        for pickle_attr in self.pickle_attrs:
            setattr(
                self,
                pickle_attr,
                joblib.load(input_artifact.path / f"{pickle_attr}.pkl")
            )



"""
Main evaluation CLI

"""
from click import command
from microcosm.loaders import load_from_dict
from microcosm.object_graph import ObjectGraph

from microcosm_sagemaker.app_hooks import create_evaluate_app
from microcosm_sagemaker.commands.options import input_artifact_option, input_data_option
from microcosm_sagemaker.input_data import InputData


@command()
@input_data_option()
@input_artifact_option()
def main(input_data, input_artifact):
    graph = create_evaluate_app(
        extra_loader=load_from_dict(
            root_input_artifact_path=input_artifact.path,
        ),
    )

    run_evaluate(
        graph=graph,
        input_data=input_data,
    )


def run_evaluate(
    graph: ObjectGraph,
    input_data: InputData,
) -> None:

    graph.experiment_metrics.init()

    graph.active_evaluation(graph.active_bundle, input_data)




import contextlib
import os
from pathlib import Path
from typing import Union


class InputData:
    def __init__(self, path: Union[str, Path]):
        self.path = Path(path)

    def __repr__(self) -> str:
        return f'{type(self).__name__}("{str(self.path)}")'

    @contextlib.contextmanager
    def cd(self):
        """
        Change working directory to directory of input data.

        """
        old_path = os.getcwd()
        os.chdir(self.path)
        try:
            yield
        finally:
            os.chdir(old_path)
